

#include "pubXmrAddress.h"

#include "keccakHash.h"

#include "base58.h"
#include "bigintutil.h"
#include "util.h"




XmrPubKey PubXmrAddress::getSpendKey() const
{
	XmrEllipticCurve ec;
	const XmrECPoint point = ec.recoverPoint(pubSpendVal);
	const XmrPubKey pubkey(point);
	return pubkey; //point;
}


XmrPubKey PubXmrAddress::getViewKey() const
{
	XmrEllipticCurve ec;
	const XmrECPoint point = ec.recoverPoint(pubViewVal);
	const XmrPubKey pubkey(point);
	return pubkey; //return point;
}






PubXmrAddress::PubXmrAddress(const BigInt<257>& inPubSpend, const BigInt<257>& inPubView, const bool inIsSubaddress) : 
	pubSpendVal(inPubSpend), pubViewVal(inPubView), isSubaddress(inIsSubaddress)
{
	//Todo: Check validity
}


PubXmrAddress::PubXmrAddress(const XmrPubKey& inPubSpend, const XmrPubKey& inPubView, const bool inIsSubaddress) : 
	pubSpendVal(inPubSpend.xyVal()), pubViewVal(inPubView.xyVal()), isSubaddress(inIsSubaddress)
{
	//Todo: Check validity
}





PubXmrAddress::PubXmrAddress(const std::string& pubAddr)
{	
	if(pubAddr.size() != 95)
	{
		throw std::string("Xmr address must be exactly 95 characters");
	}
	

	BigInt<600> totResult;

	//Split in 11 byte blocks
	
	for(int i = 0 ; ; i++)
	{
		if(i*11 > pubAddr.size())
		{
			break;
		}
		
		const std::string first = pubAddr.substr(i*11, 11);
		//const std::string firstRev = doreverseOrder(first);
			
		const BigInt<600> result = Base58Decode<600>(first);

		if(first.size() < 11)
		{
			totResult <<= 40;
			totResult += result;			
		}
		else
		{
			totResult <<= 64;
			totResult += result;
		}		
	}


	const BigInt<600> network = totResult >> 544;
	const int netVal = network.intVal();

	if(netVal != 18 && netVal != 42)
	{
		throw std::string("Error, only xmr main chain supported");
	}

	isSubaddress = (netVal == 42) ? true : false;

	const BigInt<600> pubSpend = (totResult >> 288) & 256;
	const BigInt<257> myPubSpendVal = reverseIntBytes<600>(pubSpend);

	pubSpendVal = myPubSpendVal;
	
	//std::cout << "pubSpendVal: " << pubSpendVal << std::endl;
	
	const BigInt<600> pubView = (totResult >> 32) & 256;
	const BigInt<257> myPubViewVal = reverseIntBytes<600>(pubView);	
	//const BigInt<257> myPubViewVal(pubView);
	pubViewVal = myPubViewVal;
	//std::cout << "pubViewVal : " << pubViewVal << std::endl;
	//std::cout << std::endl;


	const BigInt<600> checksum = (totResult >> 0) & 32;

	const BigInt<600> forChecksumComputation = (totResult >> 32) & 520;

	const std::string binBuf = forChecksumComputation.getBinary();

	if(binBuf.size() != 65)
	{
		throw std::string("Internal error, buffer for checksum computation should be 65 bytes");
	}

	const std::string r = computeKeccak1600(binBuf);

	const std::string computedChecksum = r.substr(0, 4);

	const BigInt<600> computedChecksumInt(BigInt<600>::fromBytes(computedChecksum));

	if(checksum != computedChecksumInt)
	{
		std::cout << "Given checksum   : " << checksum << std::endl;
		std::cout << "Computed checksum: " << computedChecksumInt << std::endl;
		
		throw std::string("Checksum error");
	}	
}




std::string prefixOnes(std::string str, const int num)
{
	while(str.size() < num)
	{
		str = std::string("1") + str;
	}
	
	return str;
}


std::string xmrBase58(const std::string& buf)
{
	std::string result;
	
	for(int i = 0 ; i < 9 ; i++)
	{
		//std::cout << "i: " << i << std::endl;

		if(i == 8)
		{
			const std::string part = buf.substr(i * 8, 5);
			const BigInt<41> val(BigInt<41>::fromBytes(part));
	
			//std::cout << "To be encoded: " << val << std::endl;
	
			const std::string encodedPart = Base58Encode<41>(val);	
			const std::string encoded = prefixOnes(encodedPart, 7);
			//std::cout << "encoded: " << encoded << std::endl;
			result += encoded;	
			
		}
		else
		{
			const std::string part = buf.substr(i * 8, 8);
			const BigInt<65> val(BigInt<65>::fromBytes(part));
	
			//std::cout << "To be encoded: " << val << std::endl;
	
			const std::string encodedPart = Base58Encode<65>(val);				
			const std::string encoded = prefixOnes(encodedPart, 11);
			//std::cout << "encoded: " << encoded << std::endl;


			result += encoded;	
		}
	
	}
	
	
	
	return result;
}




std::string PubXmrAddress::toString() const
{
	const int networkNum = isSubaddress ? 42 : 18;
	
	const BigInt<8> network(networkNum);
	
	std::string result;
	
	const std::string netStr = network.getBinary();
	const std::string netStrByte = netStr.substr(netStr.size() - 1);
		
	const std::string spendValRaw = pubSpendVal.getBinaryLong();	
	//std::cout << "spendValRaw: " << hexString(spendValRaw) << std::endl;
	
	const std::string spendVal = reverseStr(spendValRaw);	
	//std::cout << "spendVal     : " << hexString(spendVal) << std::endl;
	
	const std::string trimedSval = spendVal.substr(0, 32); //spendVal.size() - 32);
	//std::cout << "trimedSval   : " << hexString(trimedSval) << std::endl;
	
	const std::string viewValRaw = pubViewVal.getBinaryLong();	
	const std::string viewVal = reverseStr(viewValRaw);	
	//std::cout << "viewVal        : " << hexString(viewVal) << std::endl;


	const std::string trimedVval = viewVal.substr(0, 32); //viewVal.size() - 32);
	//std::cout << "trimedVval     : " << hexString(trimedVval) << std::endl;

	//std::cout << std::endl;
	const std::string buf = netStrByte + trimedSval + trimedVval;

	//std::cout << "buf: " << hexString(buf) << std::endl;

	if(buf.size() != 65)
	{
		throw std::string("Internal error, the buffer for checksum computation should be exactly 65 bytes");
	}

	//const std::string bufRev = doreverseOrder(buf);	

	//std::cout << "SIZE OF BUF TO COMPUTE KECCAC FOR: " << buf.size() << std::endl;

	//std::cout << "WTILL COMPUTE THE KECCAC FOR: " << hexString(buf) << std::endl;

	const std::string keccakHash = computeKeccak1600(buf);
	const std::string computedChecksum = keccakHash.substr(0, 4);
	const std::string totalBuf = buf + computedChecksum;
	const std::string finalAddress = xmrBase58(totalBuf);

	return finalAddress;
}




